@@ -6,8 +6,14 @@ is0  <- function(x) !is.na(x) & x == 0
 isN0 <- function(x)  is.na(x) | x != 0
 all0 <- function(x) !any(is.na(x)) && all(x == 0)
 
-allTrue  <- function(x) !any(is.na(x)) && all(x)
-allFalse <- function(x) !any(is.na(x)) && !any(x)
+allTrue  <- function(x) all(x)  && !any(is.na(x))
+allFalse <- function(x) !any(x) && !any(is.na(x))
+
+as1 <- function(x, mod=mode(x))
+    switch(mod, "integer" = 1L, "numeric" = 1, "logical" = TRUE, "complex" = 1+0i)
+as0 <- function(x, mod=mode(x))
+    switch(mod, "integer" = 0L, "numeric" = 0, "logical" = FALSE, "complex" = 0+0i)
+
 
 ## maybe we should have this in base, maybe via an .Internal(paste0(.)) -> do_paste(.. op=2)
 paste0 <- function(...) paste(..., sep = '')
@@ -32,22 +38,125 @@ paste0 <- function(...) paste(..., sep = '')
 		  fun, cl1, cl2), call. = FALSE)
 }
 
+## we can set this to FALSE and possibly measure speedup:
+.copyClass.check <- TRUE
+
 ## This should be done in C and be exported by 'methods':  [FIXME - ask JMC ]
 copyClass <- function(x, newCl, sNames =
-		      intersect(slotNames(newCl), slotNames(x))) {
+		      intersect(slotNames(newCl), slotNames(x)),
+                      check = .copyClass.check)
+{
     r <- new(newCl)
-    for(n in sNames)
-	slot(r, n) <- slot(x, n)
+    ## Equivalent of
+    ##   for(n in sNames) slot(r, n, check=check) <- slot(x, n)  :
+    if(check) for(n in sNames) slot(r, n) <- slot(x, n)
+    else for(n in sNames) # don't check, be fast
+	.Call("R_set_slot", r, n, slot(x,n), PACKAGE = "methods")
     r
 }
 
+
+attrSlotNames <- function(m) {
+    ## slotnames of Matrix objects which *not* directly content related
+    sn <- slotNames(m); sn[!(sn %in% c("x","i","j","p"))]
+}
+
+##' @param m
+##' @return the slots of 'm' which are "attributes" of them kind.
+attrSlots <- function(m) sapply(attrSlotNames(m), function(sn) slot(m, sn),
+				simplify = FALSE)
+
+attr.all_Mat <- function(target, current, check.attributes = TRUE, ...) {
+    msg <- if(check.attributes)
+	attr.all.equal(attrSlots(target), attrSlots(current), ...)
+    if((c1 <- class(target)) != (c2 <- class(current)))
+
+	## list(): so we can easily check for this
+	list(c(msg, paste("class(target) is ", class(target), ", current is ",
+				 class(current), sep = "")))
+    else msg
+}
+
+
+
 ## chol() via "dpoMatrix"
-cholMat <- function(x, pivot, ...) {
-    px <- as(x, "dpoMatrix")
-    if (isTRUE(validObject(px, test=TRUE))) chol(px)
+cholMat <- function(x, pivot = FALSE, ...) {
+    ## This will only be called for *dense* matrices
+    px <- as(x, if(length(x@x) < prod(dim(x))) ## packed
+	     "dppMatrix" else "dpoMatrix")
+    if (isTRUE(validObject(px, test=TRUE))) chol(px, pivot, ...)
     else stop("'x' is not positive definite -- chol() undefined.")
 }
 
+##  sign( <permutation> ) == determinant( <pMatrix>)
+
+signPerm <- function(p)
+{
+    ## Purpose: sign(<permutation>) via the cycles
+    ## ----------------------------------------------------------------------
+    ## Arguments: a permutation of 1:n
+    ## ----------------------------------------------------------------------
+    ## Author: Peter Dalgaard, 14 Apr 2008 // speedup: Martin Maechler 2008-04-16
+
+    n <- length(p)
+    x <- integer(n)
+    ii <- seq_len(n)
+    for (i in ii) {
+	z <- ii[!x][1]             # index of first unmarked x[] entry
+	if (is.na(z)) break
+	repeat { ## mark x[] <- i  for those in i-th cycle
+	    x[z] <- i
+	    z <- p[z]
+	    if (x[z]) break
+	}
+    }
+    ## Now, table(x) gives the cycle lengths,
+    ## where  split(seq_len(n), x)  would give the cycles list
+    ## tabulate(x, i - 1L) is quite a bit faster than the equivalent
+    ## table(x)
+    clen <- tabulate(x, i - 1L)
+    ## The sign is -1 (<==>  permutation is odd)  iff
+    ## the cycle factorization contains an odd number of even-length cycles:
+    1L - (sum(clen %% 2 == 0) %% 2L)*2L
+}
+
+
+detSparseLU <- function(x, logarithm = TRUE, ...) {
+    ## Purpose: Compute determinant() from  lu.x = lu(x)
+    ## ----------------------------------------------------------------------
+    ## Author: Martin Maechler, Date: 15 Apr 2008
+
+    if(any(x@Dim == 0)) return(mkDet(numeric(0)))
+    ll <- tryCatch(lu(x), error = function(e)e)
+
+    if(inherits(ll, "error")) {
+	## LU-decomposition failed:
+        if(length(grep("singular", ll$message, fixed=TRUE)))
+            ## <== Matrix singular and we behave as if "==>" was sure :
+            return(mkDet(ldet=-Inf, logarithm=logarithm, sig = 1L))
+        else stop(ll$message, call. = FALSE)
+    }
+    ## else
+    stopifnot(all(c("L","U") %in% slotNames(ll))) # ensure we have *sparse* LU
+    r <- mkDet(diag(ll@U), logarithm)
+    ## Det(x) == Det(P L U Q) == Det(P) * 1 * Det(U) * Det(Q); where Det(P), Det(Q) in {-1,1}
+    r$sign <- r$sign * signPerm(ll@p + 1L) * signPerm(ll@q + 1L)
+    r
+}
+
+
+## Log(Determinant) from diagonal ... used several times
+
+mkDet <- function(d, logarithm = TRUE, ldet = sum(log(abs(d))),
+                  sig = -1L+2L*as.integer(prod(sign(d)) >= 0))
+{		# sig: -1 or +1 (not 0 !)
+    modulus <- if (logarithm) ldet else exp(ldet)
+    attr(modulus, "logarithm") <- logarithm
+    val <- list(modulus = modulus, sign = sig)
+    class(val) <- "det"
+    val
+}
+
 dimCheck <- function(a, b) {
     da <- dim(a)
     db <- dim(b)
@@ -69,6 +178,7 @@ mmultCheck <- function(a, b) {
 }
 
 dimNamesCheck <- function(a, b) {
+    ## Constructs "sensical" dimnames for something like  a + b ;
     ## assume dimCheck() has happened before
     nullDN <- list(NULL,NULL)
     h.a <- !identical(nullDN, dna <- dimnames(a))
@@ -78,11 +188,10 @@ dimNamesCheck <- function(a, b) {
 	else if(!h.a) dnb
 	else { ## both have non-trivial dimnames
 	    r <- dna # "default" result
-	    for(j in 1:2) {
-		dn <- dnb[[j]]
+	    for(j in 1:2) if(!is.null(dn <- dnb[[j]])) {
 		if(is.null(r[[j]]))
 		    r[[j]] <- dn
-		else if (!is.null(dn) && any(r[[j]] != dn))
+		else if(!identical(r[[j]], dn))
 		    warning(gettextf("dimnames [%d] mismatch in %s", j,
 				     deparse(sys.call(sys.parent()))),
 			    call. = FALSE)
@@ -116,6 +225,13 @@ colCheck <- function(a, b) {
     da[2]
 }
 
+## used for is.na(<nsparse>)  but not only:
+is.na_nsp <- function(x) {
+    d <- x@Dim
+    new(if(d[1] == d[2]) "nsCMatrix" else "ngCMatrix",
+        Dim = d, Dimnames = x@Dimnames, p = rep.int(0L, d[2]+1L))
+}
+
 ## Note: !isPacked(.)  i.e. `full' still contains
 ## ----  "*sy" and "*tr" which have "undefined" lower or upper part
 isPacked <- function(x)
@@ -143,21 +259,49 @@ emptyColnames <- function(x, msg.if.not.empty = FALSE)
     x
 }
 
+idiag <- function(n, p=n)
+{
+    ## Purpose: diag() returning  *integer*
+    ## --------------------------------------------------------
+    ## Author: Martin Maechler, Date:  8 Dec 2007, 23:13
+    r <- matrix(0L, n,p)
+    if ((m <- min(n, p)) > 0)
+	r[1 + 0:(m - 1) * (n + 1)] <- 1L
+    r
+}
+
+## The indices of the diagonal entries of an  n x n matrix,  n >= 1
+## i.e. indDiag(n) === which(diag(n) == 1)
+indDiag <- function(n) cumsum(c(1L, rep.int(n+1L, n-1)))
+
 ### TODO:  write in C and port to base (or 'utils') R
-indTri <- function(n, upper = TRUE) {
-    ## == which(upper.tri(diag(n)) or
-    ##	  which(lower.tri(diag(n)) -- but much more efficiently for largish 'n'
+### -----
+### "Theory" behind this: /u/maechler/R/MM/MISC/lower-tri-w.o-matrix.R
+indTri <- function(n, upper = TRUE, diag = FALSE) {
+    ## Indices of strict upper/lower triangular part
+    ## == which(upper.tri(diag(n), diag=diag) or
+    ##	  which(lower.tri(diag(n), diag=diag) -- but
+    ## more efficiently for largish 'n'
     stopifnot(length(n) == 1, n == (n. <- as.integer(n)), (n <- n.) >= 0)
-    if(n <= 2)
-	return(if(n == 2) as.integer(if(upper) n+1 else n) else integer(0))
-    ## First, compute the 'diff(.)'  fast.  Use integers
-    one <- 1:1 ; two <- 2:2
-    n1 <- n - one
-    n2 <- n1 - one
-    r <- rep.int(one, n*n1/two - one)
-    r[cumsum(if(upper) 1:n2 else c(n1, if(n >= 4) n2:two))] <- if(upper) n:3 else 3:n
-    ## now have "dliu" difference; revert to "liu":
-    cumsum(c(if(upper) n+one else two, r))
+    if(n <= 2) {
+        if(n == 0) return(integer(0))
+        if(n == 1) return(if(diag) 1L else integer(0))
+        ## else n == 2
+        v <- if(upper) 3L else 2L
+	return(if(diag) c(1L, v, 4L) else v)
+    }
+
+    ## n >= 3 [also for n == 2 && diag (==TRUE)] :
+
+    ## First, compute the 'diff(.)' of the result [fast, using integers]
+    n. <- if(diag) n else n - 1L
+    n1 <- n. - 1L
+    ## all '1' but a few
+    r <- rep.int(1L, choose(n.+1, 2) - 1)
+    tt <- if(diag) 2L else 3L
+    r[cumsum(if(upper) 1:n1 else n.:2)] <- if(upper) n:tt else tt:n
+    ## now have differences; revert to "original":
+    cumsum(c(if(diag) 1L else if(upper) n+1L else 2L, r))
 }
 
 
@@ -213,12 +357,30 @@ nz.NA <- function(x, na.value) {
     ## na.value: TRUE: NA's give TRUE, they are not 0
     ##             NA: NA's are not known ==> result := NA
     ##          FALSE: NA's give FALSE, could be 0
-    stopifnot(is.logical(na.value) && length(na.value) == 1)
+    stopifnot(is.logical(na.value), length(na.value) == 1)
     if(is.na(na.value)) x != 0
     else  if(na.value)	isN0(x)
     else		x != 0 & !is.na(x)
 }
 
+### This assumes that e.g. the i-slot in Csparse is *not* over-allocated:
+nnzSparse <- function(x, cl = class(x), cld = getClassDef(cl))
+{
+    ## Purpose: number of *stored* / structural non-zeros {NA's counted too}
+    ## ----------------------------------------------------------------------
+    ## Arguments: x sparseMatrix
+    ## ----------------------------------------------------------------------
+    ## Author: Martin Maechler, 18 Apr 2008
+    if(extends(cld, "CsparseMatrix") || extends(cld, "TsparseMatrix"))
+	length(x@i)
+    else if(extends(cld, "RsparseMatrix"))
+	length(x@j)
+    else if(extends(cld, "pMatrix"))	# is "sparse" too
+	x@Dim[1]
+    else stop("'x' must be sparseMatrix")
+}
+
+
 ## Number of "structural" non-zeros --- this is  nnzmax() in Matlab
 ##        of effectively  non-zero values =      nnz()     "   "
 
@@ -234,73 +396,121 @@ nnzero <- function(x, na.counted = NA) {
     if(!extends(cld, "Matrix"))
 	sum(nz.NA(x, na.counted))
     else { ## Matrix
-       iSym <- extends(cld, "symmetricMatrix")
-       if(extends(cld, "pMatrix")) # is "sparse" too
-	   nrow(x)
-       else if(extends(cld, "sparseMatrix")) {
-	   nn <-
-	       if(extends(cld, "nMatrix")) # <==> no 'x' slot
-		   switch(.sp.class(cl),
-			  "CsparseMatrix" = length(x@i),
-			  "TsparseMatrix" = length(x@i),
-			  "RsparseMatrix" = length(x@j))
-	       else ## consider NAs in 'x' slot:
-		   sum(nz.NA(x@x, na.counted))
-	   if(iSym) (nn+nn - sum(nz.NA(diag(x), na.counted))) else nn
-       }
-       else if(extends(cld, "diagonalMatrix"))
+	d <- x@Dim
+	if(any(d == 0)) return(0L)
+	n <- d[1]
+	iSym <- extends(cld, "symmetricMatrix")
+	if(extends(cld, "pMatrix"))	# is "sparse" too
+	    n
+	else if(extends(cld, "diagonalMatrix"))
 	    sum(nz.NA(diag(x), na.counted))
-       else { ## dense, not diagonal: Can use 'x' slot;
-           nn <- sum(nz.NA(as_geClass(x, cl)@x, na.counted))
-           if(iSym && length(x@x) < prod(dim(x))) ## packed symmetric
-               ## n(n+1)/2  |--> n^2
-               nn <- (nn + nn) - as.integer(sqrt(2*nn))
-           nn
-       }
+	else if(extends(cld, "sparseMatrix")) {
+	    nn <- switch(.sp.class(cl),
+                         "CsparseMatrix" = x@p[d[2]+1L],# == length(x@i) only if not over-alloc.
+                         "TsparseMatrix" = length(x@i),
+                         "RsparseMatrix" = x@p[n+1L])
+	    if(!extends(cld, "nMatrix")) # <==> has 'x' slot : consider NAs in it:
+		nn <- sum(nz.NA(if(nn < length(x@x)) x@x[seq_len(nn)] else x@x,
+				na.counted))
+
+	    if(iSym)
+		nn+nn - sum(nz.NA(diag(x), na.counted))
+	    else if(extends(cld, "triangularMatrix") && x@diag == "U")
+		nn + n else nn
+	}
+	else {
+	    ## dense, not diagonal: Can use 'x' slot;
+	    if(iSym || extends(cld, "triangularMatrix")) {
+		## now !iSym  <==> "triangularMatrix"
+		upper <- (x@uplo == "U")
+		if(length(x@x) < n*n) { ## packed symmetric | triangular
+		    if(iSym) {
+			## indices of *diagonal* entries for packed :
+			iDiag <- cumsum(if(upper) 1:n else c(1L, if(n > 1)n:2))
+			## symmetric packed: count off-diagonals *twice*
+			2L* sum(nz.NA(x@x[-iDiag], na.counted)) +
+			    sum(nz.NA(x@x[ iDiag], na.counted))
+		    }
+		    else ## triangular packed
+			sum(nz.NA(x@x, na.counted))
+		}
+		else {
+		    ## not packed, but may have "arbitrary"
+		    ## entries in the non-relevant upper/lower triangle
+		    s <- sum(nz.NA(x@x[indTri(n, upper=upper)], na.counted))
+		    (if(iSym) 2L * s else s) +
+			(if(!iSym && x@diag == "U")
+			 n else sum(nz.NA(x@x[indDiag(n)], na.counted)))
+		}
+	    }
+	    else { ## dense general <--> .geMatrix
+		sum(nz.NA(x@x, na.counted))
+	    }
+	}
     }
 }
 
 ## For sparseness handling, return a
 ## 2-column (i,j) matrix of 0-based indices of non-zero entries:
-non0ind <- function(x, classDef.x = getClassDef(class(x)))
+
+non0.i <- function(M, cM = class(M), uniqT=TRUE) {
+    if(extends(cM, "TsparseMatrix")) {
+	if(uniqT && is_not_uniqT(M))
+	    .Call(compressed_non_0_ij, as(M,"CsparseMatrix"), TRUE)
+	else cbind(M@i, M@j)
+    } else if(extends(cM, "pMatrix")) {
+	cbind(seq_len(nrow(M)), M@perm) - 1L
+    } else { ## C* or R*
+	isC <- extends(cM, "CsparseMatrix")
+	.Call(compressed_non_0_ij, M, isC)
+    }
+}
+
+non0ind <- function(x, classDef.x = getClassDef(class(x)),
+		    uniqT = TRUE, xtendSymm = TRUE)
 {
     if(is.numeric(x))
 	return(if((n <- length(x))) (0:(n-1))[isN0(x)] else integer(0))
     ## else
-
     stopifnot(extends(classDef.x, "sparseMatrix"))
 
-    non0.i <- function(M, cM = class(M)) {
-	if(extends(cM, "TsparseMatrix"))
-	    return(unique(cbind(M@i,M@j)))
-	if(extends(cM, "pMatrix"))
-	    return(cbind(seq_len(nrow(M)), M@perm) - 1:1)
-	## else: C* or R*
-	isC <- extends(cM, "CsparseMatrix")
-	.Call(compressed_non_0_ij, M, isC)
-    }
-
-    if(extends(classDef.x, "symmetricMatrix")) { # also get "other" triangle
-	ij <- non0.i(x, classDef.x)
+    ij <- non0.i(x, classDef.x, uniqT=uniqT)
+    if(xtendSymm && extends(classDef.x, "symmetricMatrix")) { # also get "other" triangle
 	notdiag <- ij[,1] != ij[,2]# but not the diagonals again
 	rbind(ij, ij[notdiag, 2:1])
     }
     else if(extends(classDef.x, "triangularMatrix")) { # check for "U" diag
 	if(x@diag == "U") {
-	    i <- seq_len(dim(x)[1]) - 1:1
-	    rbind(non0.i(x, classDef.x), cbind(i,i))
-	} else non0.i(x, classDef.x)
+	    i <- seq_len(dim(x)[1]) - 1L
+	    rbind(ij, cbind(i,i))
+	} else ij
     }
     else
-	non0.i(x, classDef.x)
+	ij
 }
 
-## nr= nrow: since  i in {0,1,.., nrow-1}  these are 1:1 "decimal" encodings:
-## Further, these map to and from the usual "Fortran-indexing" (but 0-based)
-encodeInd  <- function(ij,  nr) ij[,1] + ij[,2] * nr
-encodeInd2 <- function(i,j, nr) i      +  j     * nr
+if(FALSE) { ## -- now have  .Call(m_encodeInd, ...) etc :
 
-decodeInd <- function(code, nr) cbind(code %% nr, code %/% nr)
+## nr= nrow: since  i in {0,1,.., nrow-1}  these are 1L "decimal" encodings:
+## Further, these map to and from the usual "Fortran-indexing" (but 0-based)
+## __ care against integer overflow __
+encodeInd  <- function(ij, di) {
+    stopifnot(length(di) == 2)
+    nr <- di[1L]
+    if(prod(di) >= .Machine$integer.max) nr <- as.double(nr)
+    ij[,1] + ij[,2] * nr
+}
+encodeInd2 <- function(i,j, di) {
+    stopifnot(length(di) == 2)
+    nr <- di[1L]
+    if(prod(di) >= .Machine$integer.max) nr <- as.double(nr)
+    i +  j * nr
+}
+}## no more needed
+
+## 'code' is 0-based; as.integer(.) because encode*() may produce double
+decodeInd <- function(code, nr) cbind(as.integer(code %% nr),
+				      as.integer(code %/% nr))
 
 complementInd <- function(ij, dim)
 {
@@ -308,23 +518,22 @@ complementInd <- function(ij, dim)
     ##		but as 1-based indices
     n <- prod(dim)
     if(n == 0) return(integer(0))
-    ii <- 1:n
-    ii[-(1 + encodeInd(ij, nr = dim[1]))]
+    seq_len(n)[-(1L + .Call(m_encodeInd, ij, dim))]
 }
 
 unionInd <- function(ij1, ij2) unique(rbind(ij1, ij2))
 
-intersectInd <- function(ij1, ij2, nrow) {
+intersectInd <- function(ij1, ij2, di) {
     ## from 2-column (i,j) matrices where i in {0,.., nrow-1},
     ## return only the *common* entries
-    decodeInd(intersect(encodeInd(ij1, nrow),
-			encodeInd(ij2, nrow)), nrow)
+    decodeInd(intersect(.Call(m_encodeInd, ij1, di),
+			.Call(m_encodeInd, ij2, di)), nr=di[1])
 }
 
-WhichintersectInd <- function(ij1, ij2, nrow) {
+WhichintersectInd <- function(ij1, ij2, di) {
     ## from 2-column (i,j) matrices where i \in {0,.., nrow-1},
     ## find *where*  common entries are in ij1 & ij2
-    m1 <- match(encodeInd(ij1, nrow), encodeInd(ij2, nrow))
+    m1 <- match(.Call(m_encodeInd, ij1, di), .Call(m_encodeInd, ij2, di))
     ni <- !is.na(m1)
     list(which(ni), m1[ni])
 }
@@ -362,14 +571,13 @@ uniqTsparse <- function(x, class.x = c(class(x))) {
 ## would be slightly more efficient than as( <dgC> , "dgTMatrix")
 ## but really efficient would be to use only one .Call(.) for uniq(.) !
 
-drop0 <- function(x, clx = c(class(x))) {
-    if(!extends(clx, "CsparseMatrix"))
-        clx <- sub(".Matrix$", "CMatrix", clx)
-    ## FIXME: Csparse_drop should do this, but it
-    ##	      drops triangularity and symmetry :
-    ## .Call(Csparse_drop, as_CspClass(x, clx), 0)
-    as_CspClass(.Call(Csparse_drop, as_CspClass(x, clx), 0.),
-		clx)
+## Eearlier was:
+## drop0 <- function(x, clx = c(class(x)), tol = 0) {
+drop0 <- function(x, tol = 0, is.Csparse = NA) {
+    .Call(Csparse_drop,
+	  if(isTRUE(is.Csparse) || is.na(is.Csparse) && is(x, "CsparseMatrix")) x else
+          as(x, "CsparseMatrix"),
+	  tol)
 }
 
 uniq <- function(x) {
@@ -382,12 +590,12 @@ asTuniq <- function(x) {
 }
 
 ## is 'x' a uniq Tsparse Matrix ?
-is_not_uniqT <- function(x, nr = nrow(x))
-    is.unsorted(x@j) || any(duplicated(encodeInd2(x@i, x@j, nr)))
+is_not_uniqT <- function(x, di = dim(x))
+    is.unsorted(x@j) || any(duplicated(.Call(m_encodeInd2, x@i, x@j, di)))
 
 ## is 'x' a TsparseMatrix with no duplicated entries (to be *added* for uniq):
-is_duplicatedT <- function(x, nr = nrow(x))
-    any(duplicated(encodeInd2(x@i, x@j, nr)))
+is_duplicatedT <- function(x, di = dim(x))
+    any(duplicated(.Call(m_encodeInd2, x@i, x@j, di)))
 
 
 if(FALSE) ## try an "efficient" version
@@ -495,9 +703,12 @@ tT2gT <- function(x, cl = class(x),
 	new(toClass, Dim = d, Dimnames = x@Dimnames,
 	    i = c(x@i, if(uDiag) 0:(n-1)),
 	    j = c(x@j, if(uDiag) 0:(n-1)),
-	    x = c(x@x, if(uDiag) rep.int(1,n)))
+	    x = c(x@x, if(uDiag) rep.int(if(extends(cld, "dMatrix")) 1 else TRUE, n)))
 }
 
+## Fast very special one
+## .gT2tC <- function(x, uplo, diag) .Call(Tsparse_to_tCsparse, x, uplo, diag)
+
 gT2tT <- function(x, uplo, diag,
 		  cl = class(x),
 		  toClass = paste(substr(cl,1,1), "tTMatrix", sep=''),# d,l,i,z
@@ -602,16 +813,26 @@ l2d_meth <- function(x) {
     }
 }
 
+## a faster simpler version [for sparse matrices, i.e., never diagonal]
+.M.shapeC <- function(x, clx = class(x)) {
+    if(is.character(clx)) # < speedup: get it once
+	clx <- getClassDef(clx)
+    if	   (extends(clx, "triangularMatrix")) "t"
+    else if(extends(clx, "symmetricMatrix"))  "s" else "g"
+}
+
 
 class2 <- function(cl, kind = "l", do.sub = TRUE) {
     ## Find "corresponding" class; since pos.def. matrices have no pendant:
-    if	   (cl == "dpoMatrix") paste(kind, "syMatrix", sep='')
-    else if(cl == "dppMatrix") paste(kind, "spMatrix", sep='')
+    if(cl %in% c("dpoMatrix","corMatrix"))
+	paste(kind, "syMatrix", sep='')
+    else if(cl == "dppMatrix")
+	paste(kind, "spMatrix", sep='')
     else if(do.sub) sub("^[a-z]", kind, cl)
     else cl
 }
 
-## see also as_geClass() below
+## see also as_smartClass() below
 geClass <- function(x) {
     if     (is(x, "dMatrix")) "dgeMatrix"
     else if(is(x, "lMatrix")) "lgeMatrix"
@@ -621,7 +842,7 @@ geClass <- function(x) {
 	      class(x))
 }
 
-.dense.prefixes <- c("d" = "di",
+.dense.prefixes <- c("d" = "tr", ## map diagonal to triangular
                      "t" = "tr",
                      "s" = "sy",
                      "g" = "ge")
@@ -636,6 +857,8 @@ as_dense <- function(x, cld = if(isS4(x)) getClassDef(class(x))) {
     as(x, paste(.M.kind(x, cld), .dense.prefixes[.M.shape(x, cld)], "Matrix", sep=''))
 }
 
+## This is "general" but slower than the next definition
+if(FALSE)
 .sp.class <- function(x) { ## find and return the "sparseness class"
     if(!is.character(x)) x <- class(x)
     for(cl in paste(c("C","T","R"), "sparseMatrix", sep=''))
@@ -645,10 +868,18 @@ as_dense <- function(x, cld = if(isS4(x)) getClassDef(class(x))) {
     as.character(NA)
 }
 
+.sp.class <- function(x) { ## find and return the "sparseness class"
+    if(!is.character(x)) x <- c(class(x))
+    if(any((ch <- substr(x,3,3)) == c("C","T","R")))
+        return(paste(ch, "sparseMatrix", sep=''))
+    ## else
+    NA_character_
+}
+
 
 ### Goal: Eventually get rid of these --- want to foster coercions
-### ----  *to* virtual classes whenever possible, i.e.
-##  as(*, "CsparseMatrix"),  etc
+### ----  *to* virtual classes whenever possible, e.g. as(*, "CsparseMatrix")
+## 2007-12: better goal: use them only for "matrix" [maybe speed them up later]
 
 ## Here, getting the class definition and passing it, should be faster
 as_Csparse <- function(x, cld = if(isS4(x)) getClassDef(class(x))) {
@@ -656,6 +887,7 @@ as_Csparse <- function(x, cld = if(isS4(x)) getClassDef(class(x))) {
                 .sparse.prefixes[.M.shape(x, cld)], "CMatrix", sep=''))
 }
 
+if(FALSE) # replaced by .Call(dense_to_Csparse, *) which is perfect for "matrix"
 as_Csparse2 <- function(x, cld = if(isS4(x)) getClassDef(class(x))) {
     ## Csparse + U2N when needed
     sh <- .M.shape(x, cld)
@@ -663,6 +895,13 @@ as_Csparse2 <- function(x, cld = if(isS4(x)) getClassDef(class(x))) {
     if(sh == "t") .Call(Csparse_diagU2N, x) else x
 }
 
+## *do* use this where applicable
+as_Csp2 <- function(x) {
+    ## Csparse + U2N when needed
+    x <- as(x, "CsparseMatrix")
+    if(is(x, "triangularMatrix")) .Call(Csparse_diagU2N, x) else x
+}
+
 
 ## 'cl'   : class() *or* class definition of from
 as_gCsimpl2 <- function(from, cl = class(from))
@@ -678,6 +917,7 @@ as_Sp <- function(from, shape, cl = class(from)) {
 		   if(extends(cl, "TsparseMatrix")) "TMatrix" else "CMatrix",
 		   sep=''))
 }
+## These are used in ./sparseMatrix.R:
 as_gSparse <- function(from) as_Sp(from, "g", getClassDef(class(from)))
 as_sSparse <- function(from) as_Sp(from, "s", getClassDef(class(from)))
 as_tSparse <- function(from) as_Sp(from, "t", getClassDef(class(from)))
@@ -687,14 +927,15 @@ as_geSimpl2 <- function(from, cl = class(from))
 ## to be used directly in setAs(.) needs one-argument-only  (from) :
 as_geSimpl <- function(from) as(from, paste(.M.kind(from), "geMatrix", sep=''))
 
-## smarter, (but sometimes too smart!) compared to geClass() above:
-as_geClass <- function(x, cl) {
+## Smarter, (but sometimes too smart!) compared to geClass() above:
+as_smartClass <- function(x, cl, cld = getClassDef(cl)) {
     if(missing(cl)) return(as_geSimpl(x))
     ## else
-    cld <- getClassDef(cl)
     if(extends(cld, "diagonalMatrix")  && isDiagonal(x))
+        ## diagonal* result:
 	as(x, cl)
     else if(extends(cld, "symmetricMatrix") &&  isSymmetric(x)) {
+        ## symmetric* result:
         kind <- .M.kind(x, cld)
 	as(x, class2(cl, kind, do.sub= kind != "d"))
     } else if(extends(cld, "triangularMatrix") && isTriangular(x))
@@ -704,15 +945,25 @@ as_geClass <- function(x, cl) {
 }
 
 as_CspClass <- function(x, cl) {
-    if (## diagonal is *not* sparse:
-	##(extends(cl, "diagonalMatrix") && isDiagonal(x)) ||
-	(extends(cl, "symmetricMatrix") && isSymmetric(x)) ||
-	(extends(cl, "triangularMatrix")&& isTriangular(x)))
+    ## NOTE: diagonal is *not* sparse:
+    cld <- getClassDef(cl)
+    ##(extends(cld, "diagonalMatrix") && isDiagonal(x)) ||
+    if (extends(cld, "symmetricMatrix") && isSymmetric(x))
+        forceSymmetric(as(x,"CsparseMatrix"))
+    else if (extends(cld, "triangularMatrix") && (iT <- isTriangular(x)))
 	as(x, cl)
     else if(is(x, "CsparseMatrix")) x
-    else as(x, paste(.M.kind(x, cl), "gCMatrix", sep=''))
+    else as(x, paste(.M.kind(x, cld), "gCMatrix", sep=''))
 }
 
+asTri <- function(from, newclass) {
+    ## TODO: also check for unit-diagonal: 'diag = "U"'
+    isTri <- isTriangular(from)
+    if(isTri)
+	new(newclass, x = from@x, Dim = from@Dim,
+	    Dimnames = from@Dimnames, uplo = attr(isTri, "kind"))
+    else stop("not a triangular matrix")
+}
 
 
 
@@ -733,50 +984,79 @@ isTriMat <- function(object, upper = NA) {
     ## pretest: is it square?
     d <- dim(object)
     if(d[1] != d[2]) return(FALSE)
+    TRUE.U <- structure(TRUE, kind = "U")
+    if(d[1] == 0) return(TRUE.U)
     ## else slower test
+    TRUE.L <- structure(TRUE, kind = "L")
     if(!is.matrix(object))
 	object <- as(object,"matrix")
     if(is.na(upper)) {
 	if(all0(object[lower.tri(object)]))
-	    structure(TRUE, kind = "U")
+	    TRUE.U
 	else if(all0(object[upper.tri(object)]))
-	    structure(TRUE, kind = "L")
+	    TRUE.L
 	else FALSE
     } else if(upper)
-	all0(object[lower.tri(object)])
+	if(all0(object[lower.tri(object)])) TRUE.U else FALSE
     else ## upper is FALSE
-	all0(object[upper.tri(object)])
+	if(all0(object[upper.tri(object)])) TRUE.L else FALSE
+}
+
+## For Tsparse matrices:
+isTriT <- function(object, upper = NA) {
+    ## pretest: is it square?
+    d <- dim(object)
+    if(d[1] != d[2]) return(FALSE)
+    ## else
+    TRUE.U <- structure(TRUE, kind = "U")
+    if(d[1] == 0) return(TRUE.U)
+    TRUE.L <- structure(TRUE, kind = "L")
+    if(is.na(upper)) {
+	if(all(object@i <= object@j))
+	    TRUE.U
+	else if(all(object@i >= object@j))
+	    TRUE.L
+	else FALSE
+    } else if(upper) {
+	if(all(object@i <= object@j)) TRUE.U else FALSE
+    } else { ## 'lower'
+	if(all(object@i >= object@j)) TRUE.L else FALSE
+    }
 }
 
 ## For Csparse matrices
-isTriC <- function(x, upper = NA) {
+isTriC <- function(object, upper = NA) {
     ## pretest: is it square?
-    d <- dim(x)
+    d <- dim(object)
     if(d[1] != d[2]) return(FALSE)
     ## else
-    if(d[1] == 0) return(TRUE)
+    TRUE.U <- structure(TRUE, kind = "U")
+    if(d[1] == 0) return(TRUE.U)
+    TRUE.L <- structure(TRUE, kind = "L")
     ni <- 1:d[2]
     ## the row indices split according to column:
-    ilist <- split(x@i, factor(rep.int(ni, diff(x@p)), levels= ni))
+    ilist <- split(object@i, factor(rep.int(ni, diff(object@p)), levels= ni))
     lil <- unlist(lapply(ilist, length), use.names = FALSE)
     if(any(lil == 0)) {
 	pos <- lil > 0
 	if(!any(pos)) ## matrix of all 0's
-	    return(TRUE)
+	    return(TRUE.U)
 	ilist <- ilist[pos]
 	ni <- ni[pos]
     }
-    ni0 <- ni - 1:1 # '0-based ni'
+    ni0 <- ni - 1L # '0-based ni'
     if(is.na(upper)) {
 	if(all(sapply(ilist, max, USE.NAMES = FALSE) <= ni0))
-	    structure(TRUE, kind = "U")
+	    TRUE.U
 	else if(all(sapply(ilist, min, USE.NAMES = FALSE) >= ni0))
-	    structure(TRUE, kind = "L")
+	    TRUE.L
 	else FALSE
     } else if(upper) {
-	all(sapply(ilist, max, USE.NAMES = FALSE) <= ni0)
+	if(all(sapply(ilist, max, USE.NAMES = FALSE) <= ni0))
+	    TRUE.U else FALSE
     } else { ## 'lower'
-	all(sapply(ilist, min, USE.NAMES = FALSE) >= ni0)
+	if(all(sapply(ilist, min, USE.NAMES = FALSE) >= ni0))
+	    TRUE.L else FALSE
     }
 }
 
@@ -809,17 +1089,36 @@ isTriC <- function(x, upper = NA) {
 ## Purpose: Transform a *unit diagonal* sparse triangular matrix
 ##	into one with explicit diagonal entries '1'
 
-## fast no-test version:
-.diagU2N <- function(x, cl)
+.diagU2N <- function(x, cl, checkDense = FALSE)
 {
+    ## fast "no-test" version --- we *KNOW* 'x' is 'triangularMatrix'
     if(extends(cl, "CsparseMatrix"))
 	.Call(Csparse_diagU2N, x)
+    else if(extends(cl, "TsparseMatrix"))
+	.Call(Tsparse_diagU2N, x)
     else {
 	kind <- .M.kind(x, cl)
-	xT <- as(x, paste(kind, "gTMatrix", sep=''))
-	## leave it as	T* - the caller can always coerce to C* if needed:
-	new(paste(kind, "tTMatrix", sep=''), x = xT@x, i = xT@i, j = xT@j,
-	    Dim = x@Dim, Dimnames = x@Dimnames, uplo = x@uplo, diag = "N")
+        if(checkDense && extends(cl,"denseMatrix")) {
+	    ## For denseMatrix, 'diag = "U"'
+	    ## means the 'x' slot can have wrong values which are
+	    ## documented to never be accessed
+	    if(isPacked(x)) { ## unpack
+		stop("not yet implemented for packed class ", class(x))
+		## FIXME
+	    }
+	    ## okay: now have  'x' slot of length n x n
+	    n <- dim(x)[1]
+	    if(n > 0)
+		x@x[1L+ (0:(n-1L))*(n+1L)] <-
+		    if(kind == "d") 1. else TRUE # even for "n..Matrix"
+	    x@diag <- "N"
+	    x
+        }
+        else { ## not dense, not [CT]sparseMatrix  ==>  Rsparse*
+	    .Call(Tsparse_diagU2N,
+		  as(as(x, paste(kind, "Matrix", sep='')), "TsparseMatrix"))
+	    ## leave it as T* - the caller can always coerce to C* if needed
+        }
     }
 }
 
@@ -830,7 +1129,11 @@ diagU2N <- function(x, cl = getClassDef(class(x)))
     else x
 }
 
-
+diagN2U <- function(x, cl = getClassDef(class(x)))
+{
+    if(extends(cl, "triangularMatrix") && x@diag == "N")
+	.Call(Csparse_diagN2U, as(x, "CsparseMatrix")) else x
+}
 
 .as.dgC.0.factors <- function(x) {
     if(!is(x, "dgCMatrix"))
@@ -887,3 +1190,19 @@ sp.rowMeans <- function(x, na.rm = FALSE, dims = 1, sparseResult = FALSE)
 			      sparseResult=sparseResult)
     sparsapply(x, 1, sum, sparseResult=sparseResult, na.rm=na.rm) / nc
 }
+
+all0Matrix <- function(n,m) {
+    ## an  all-0 matrix	 -- chose what Matrix() also gives -- "most efficiently"
+    n <- as.integer(n)
+    m <- as.integer(m)
+    new(if(n == m) "dsCMatrix" else "dgCMatrix",
+	Dim = c(n,m),
+	p = rep.int(0L, m+1L))
+}
+
+setZero <- function(x) {
+    ## all-0 matrix  from x  which must inherit from 'Matrix'
+    d <- x@Dim
+    new(if(d[1] == d[2]) "dsCMatrix" else "dgCMatrix",
+	Dim = d, Dimnames = x@Dimnames, p = rep.int(0L, d[2]+1L))
+}
